# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

sys.path.append('example5_motor3D')

import numpy as np
from math import pi
import math
from ngsolve import exp


def nuAca(s):
    par1 = 200
    par2 = 0.001
    par3 = 6
    nu0 = 1e7/(4*pi)
    nuVal = nu0-(nu0-par1) *exp( (-1)*par2*( s )**par3)
    return nuVal


bLinear  = False

mu0 = 4*pi*1e-7
nu0 = 1 / mu0
muriron = 5100
murmag = 1.086
BR = 1.28
numag = 1/(mu0 * murmag)
nu1lin = 1/(mu0 * muriron)

#param1 = 200
#param2 = 0.001
#param3 = 6;

order = 2
fes = Periodic(HCurl(mesh, order=order, dirichlet="outer", nograds=True))

F = Id(3)
Finv = Inv(F)
J = Det(F)

def Equation1a(u,v):
    if bLinear:
        return (nu1lin*( 1/J * F * curl(u))*(1/J * F * curl(v)) + 1e-1*(Finv.trans*u)*(Finv.trans*v))*J *dx(definedon=mesh.Materials("iron")) 
    else:
        return ( nuAca (sqrt( 1e-12 + InnerProduct(1/J * F*curl(u),1/J * F*curl(u)) ) )  *( 1/J * F * curl(u) )*( 1/J * F * curl(v) ) + 1e-1*(Finv.trans*u)*(Finv.trans*v))*J *dx(definedon=mesh.Materials("iron")) 

def Equation2a(u,v):
    return (1/mu0*( 1/J * F * curl(u) )*( 1/J * F * curl(v) ) + 1e-1*(Finv.trans*u)*(Finv.trans*v))*J *dx(definedon=mesh.Materials("air|airgap")) 

def Equation3a(u,v):
    return (numag*( 1/J * F * curl(u) )*( 1/J * F * curl(v) ) + 1e-1*(Finv.trans*u)*(Finv.trans*v) -numag*BR*CoefficientFunction((0.382683432365087,   0.923879532511288, 0))* (1/J * F * curl(v)) )*J *dx(definedon=mesh.Materials("magnet1")) 

def Equation4a(u,v):
    return (numag*(1/J * F * curl(u))*(1/J * F * curl(v)) + 1e-1*(Finv.trans*u)*(Finv.trans*v) -numag*BR*CoefficientFunction((-0.923879532511286,  -0.382683432365093, 0))*(1/J * F * curl(v)) )*J *dx(definedon=mesh.Materials("magnet2"))


def Equation_vol(u,v):
   return Equation1a(u,v) + Equation2a(u,v) + Equation3a(u,v) + Equation4a(u,v)
def Equation_bnd(u,v):
   return 0*u.Trace()*v.Trace()*ds

###################################################
def SolveNonlinearMinProblem(a,c,gfu,bprintinfo=True,tol=1e-11,maxits=25):
    res = gfu.vec.CreateVector()
    du  = gfu.vec.CreateVector()

    for it in range(maxits):
        if bprintinfo:
            print ("Newton iteration {:3}".format(it),end="")
            print ("energy = {:16}".format(a.Energy(gfu.vec)),end="")
    
        #solve linearized problem:
        a.Apply (gfu.vec, res)
        a.AssembleLinearization (gfu.vec)
        #inv = a.mat.Inverse(fesref.FreeDofs())
        inv = CGSolver (mat=a.mat, pre=c.mat, printrates = bprintinfo)
        du.data = inv * res
    
        #update iteration
        gfu.vec.data -= du

        #stopping criteria
        stopcritval = sqrt(abs(InnerProduct(du,res)))
        if bprintinfo:
            print ("<A u",it,", A u",it,">_{-1}^0.5 = ", stopcritval)
        if stopcritval < tol:
            break
        Redraw(blocking=True)
    return gfu

###################################################


def solvePDE():
    SolveNonlinearMinProblem(a,ca,gfu,bprintinfo=False,tol=1e-10,maxits=25)
                
    dCostdu.Assemble()
    a.AssembleLinearization(gfu.vec)
    inva = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky")
    gfp.vec.data = -inva.T * dCostdu.vec
    
    Draw(curl(gfu), mesh, 'B-Field')
    

def getNorm(x):
    return np.sqrt(np.dot(x,x)) 

def getRotMat(nn, ang):
    normN = getNorm(nn)
    n = 1/normN * nn
    R11 = n[0]*n[0]*(1-cos(ang))+cos(ang)
    R21 = n[1]*n[0]*(1-cos(ang))+n[2]*sin(ang)
    R31 = n[2]*n[0]*(1-cos(ang))-n[1]*sin(ang)
    
    R12 = n[0]*n[1]*(1-cos(ang))-n[2]*sin(ang)
    R22 = n[1]*n[1]*(1-cos(ang))+ cos(ang)
    R32 = n[2]*n[1]*(1-cos(ang))+ n[0]*sin(ang)
    
    R13 = n[0]*n[2]*(1-cos(ang))+ n[1]*sin(ang)
    R23 = n[1]*n[2]*(1-cos(ang))- n[0]*sin(ang)
    R33 = n[2]*n[2]*(1-cos(ang))+ cos(ang)
    
    R = np.array( ((R11, R12, R13), (R21, R22, R23), (R31, R32, R33)) )
    return R
    
def writeBrValues(gfsetCurrent, count):
    ########### writing objective function Br values ########
    curl_gfu = curl(gfu)
    r5 = 19.6666670000000
    r6 = 19.8333330000000
    myR = (r5+r6)/2
    nAngles = 90
    nZ = 21
    myZ = -0.5*myzlength

    deltaZ = myzlength / (nZ-1)
    p0 = np.array([myR, 0, myZ])

    zAxis = np.array([0,0,1])
    ang1 = (pi/2) / nAngles
    R1 = getRotMat(zAxis, ang1)

    fileName = outputFolder + "/Brvals_" + str(count) + ".m"
    myFile = open(fileName,"w+")
    myFile.write("Br_" + str(count) +" = [")
    myn2D = CoefficientFunction((X[0]/ sqrt( X[0]*X[0]+X[1]*X[1] ),X[1]/ sqrt( X[0]*X[0]+X[1]*X[1] ),0)) 

    mesh.UnsetDeformation()
    
    for iz in range(nZ):
        p0 = np.array([myR, 0, myZ])
        for iAngle in range(nAngles):
            mip = mesh(p0[0], p0[1], p0[2])      #mapped integration point
            
            myFile.write(str(np.dot(curl_gfu(mip), myn2D(mip))) + " ")  
            p0 =np.dot(R1,p0)
            
        myZ = myZ + deltaZ
        myFile.write("; ")
        
    mesh.SetDeformation(gfsetCurrent)
    
    myFile.write("]; \n")
    myFile.write("[X,Y] = meshgrid(1:"+str(nZ)+",1:"+str(nAngles)+"); surf(X', Y', Br_" + str(count) + ") \n")
    myFile.write("zlim([-2.5e-4 2.5e-4])")
    myFile.close()
    
    if count==0:
        myZ = -0.5*myzlength
        fileName_des = outputFolder + "/Brvals_desired.m"
        myFile_des = open(fileName_des,"w+")
        myFile_des.write("Brvals_des = [")
        for iz in range(nZ):
            p0 = np.array([myR, 0, myZ])
            for iAngle in range(nAngles):
                myFile_des.write(str( (-1)*ampZ5(p0[2])*sin(4*math.atan2(p0[1],p0[0])) ) + " ")  
                p0 =np.dot(R1,p0)
            
            myZ = myZ + deltaZ
            myFile_des.write("; ")
        myFile_des.write("]; \n")
        myFile_des.write("[X,Y] = meshgrid(1:"+str(nZ)+",1:"+str(nAngles)+"); surf(X', Y', Brvals_des) \n")
        myFile_des.write("zlim([-2.5e-4 2.5e-4])")
        myFile_des.close()
        


isComplex = False
isSurface = False

