# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

SURF = False

if not SURF:
    ############## Section 2.1 #################
    from ngsolve import *
    from netgen.geom2d import SplineGeometry

    geo = SplineGeometry()
    geo.AddCircle((0.5,0.5),0.5,bc="circle") 

    mesh = Mesh(geo.GenerateMesh(maxh=0.2))
    mesh.Curve(3) 
    
    fes = H1(mesh, order=3, dirichlet="circle") 

    u = fes.TrialFunction()     
    w = fes.TestFunction() 
    
    X = CoefficientFunction((x,y,z))
    
    L = LinearForm(fes)
    f1 = (2*X[1]*(1-X[1])+2*X[0]*(1-X[0]))
    L += f1 * w * dx

    a = BilinearForm(fes, symmetric=True)
    a += grad(u)*grad(w)*dx
    
    a.Assemble()
    L.Assemble()

    gfu = GridFunction(fes)
    gfu.vec.data = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky") * L.vec

    Draw (gfu, mesh, "state")
    
    input("End of Section 2.1. Press enter to continue.")
    ############## Section 2.2 #################
    v = Parameter(1)
    expr = 2*x*x+3*y
    dexpr = expr.Diff(x,v)
    print(expr)
    print(dexpr)
    
    u = fes.TrialFunction()  # symbolic object
    w = fes.TestFunction()   # symbolic object
    gf = GridFunction(fes)
    gf.Set(x*x*y)

    u = fes.TrialFunction()  # symbolic object
    w = fes.TestFunction()   # symbolic object
    gf = GridFunction(fes)
    gf.Set(x*x*y)

    print("Diff u w.r.t. x", u.Diff(x))
    print("Diff w w.r.t. x", w.Diff(x))
    print("Diff gf w.r.t. x", gf.Diff(x))
    
    input("End of Section 2.2. Press enter to continue.")
    ############## Section 3 #################
    
    ############## Section 3.1 #################    
    f = ((X[0]-0.5)/1.3)**2+(1.3*(X[1]-0.5))**2 - 0.5**2

    F = Id(2)               # symbolic identity matrix
    G_f = f * Det(F) * dx
    
    VEC = VectorH1(mesh, order=1, dirichlet="") #vectorial FE space of order 1  
    V = VEC.TestFunction()

    dJOmega_f = LinearForm(VEC)
    dJOmega_f += G_f.Diff(X, V) 
    dJOmega_f += G_f.Diff(F, grad(V))
    
    G_f_bnd = f * Det(F) * Norm( Inv(F).trans*specialcf.normal(2) ) * ds

    dJOmega_f_bnd = LinearForm(VEC)
    dJOmega_f_bnd+=G_f_bnd.Diff(X, V)                             # no trace needed
    dJOmega_f_bnd+=G_f_bnd.Diff(F,grad(V).Trace())               # trace needed
    
    input("End of Section 3.1. Press enter to continue.")
    ############## Section 3.2 #################
    d2JOmega_f = BilinearForm(VEC)
    W = VEC.TrialFunction()

    d2JOmega_f+=(G_f.Diff(X,W)+G_f.Diff(F,grad(W))).Diff(X,V) 
    d2JOmega_f+=(G_f.Diff(X,W)+G_f.Diff(F,grad(W))).Diff(F,grad(V)) 
    
    d2JOmega_f_bnd = BilinearForm(VEC)

    d2JOmega_f_bnd+=(G_f_bnd.Diff(X, W)+G_f_bnd.Diff(F,grad(W).Trace())).Diff(X, V)
    d2JOmega_f_bnd+=(G_f_bnd.Diff(X, W)+G_f_bnd.Diff(F,grad(W).Trace())).Diff(F,grad(V).Trace())
    
    input("End of Section 3.2. Press enter to continue.")
    ############## Section 4.2 #################
    ud = X[0]*(1-X[0])*X[1]*(1-X[1])
    def Cost(u):
        return (u-ud)**2 * Det(F) * dx

    #solve adjoint equation
    gfp = GridFunction(fes)
    dCostdu = LinearForm(fes)
    dCostdu += Cost(gfu).Diff(gfu, w)
    dCostdu.Assemble()
    gfp.vec.data = -a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky").T * dCostdu.vec

    Draw(gfp, mesh, "adjoint")
    
    def Equation(u,w):
        return ((Inv(F).trans * grad(u)) * (Inv(F).trans * grad(w)) - f*w)*Det(F)*dx
        
    G_pde = Cost(gfu) + Equation(gfu, gfp) 

    dJOmega_pde = LinearForm(VEC)
    dJOmega_pde += G_pde.Diff(X, V) 
    dJOmega_pde += G_pde.Diff(F, grad(V)) 
    
    input("End of Section 4.2. Press enter to continue.")
    ############## Section 4.3 #################
    
    X2 = FESpace([fes, fes])
    dsu, dsp = X2.TrialFunction()
    uTest, pTest = X2.TestFunction()
    gfV = GridFunction(VEC)
    gfW = GridFunction(VEC)
    gfV.Set((X[0]*X[0]*X[1]*exp(X[1]),X[1]*X[1]*X[0]*exp(X[0]))) 
    gfW.Set((X[1]*X[1]*X[0]*exp(X[0]),X[0]*X[0]*X[1]*exp(X[1]))) 
    
    shapeHessLag2 = BilinearForm(X2)
    shapeGradLag2 = LinearForm(X2)

    shapeHessLag2 += (G_pde.Diff(gfu, uTest)).Diff(gfu, dsu)   #block (1,1)
    shapeHessLag2 += (G_pde.Diff(gfu, uTest)).Diff(gfp, dsp)   #block (1,2)
    shapeHessLag2 += (G_pde.Diff(gfp, pTest)).Diff(gfu, dsu)   #block (2,1)
        
    #line 1
    shapeGradLag2 += (G_pde.Diff(gfu, uTest)).Diff(F, grad(gfV)) 
    shapeGradLag2 += (G_pde.Diff(gfu, uTest)).Diff(X, gfV) 

    #line 2
    shapeGradLag2 += (G_pde.Diff(gfp,pTest)).Diff(F, grad(gfV))
    shapeGradLag2 += (G_pde.Diff(gfp,pTest)).Diff(X, gfV) 
    
    gfCombined2 = GridFunction(X2)
    shapeHessLag2.Assemble()
    shapeGradLag2.Assemble()
    gfCombined2.vec.data = shapeHessLag2.mat.Inverse(X2.FreeDofs(), inverse = "umfpack") * shapeGradLag2.vec

    gfdsu = GridFunction(fes)
    gfdsp = GridFunction(fes)
    gfdsu.vec.data = gfCombined2.components[0].vec
    gfdsp.vec.data = gfCombined2.components[1].vec

    Draw(gfdsu, mesh, "dsu")
    Draw(gfdsp, mesh, "dsp")
    
    w1 = fes.TrialFunction()
    q1 = fes.TrialFunction()

    shapeHess11 = BilinearForm(VEC)
    shapeHess11 += (G_pde.Diff(F,grad(W))+G_pde.Diff(X, W)).Diff(F,grad(V)) 
    shapeHess11 += (G_pde.Diff(F,grad(W))+G_pde.Diff(X, W)).Diff(X, V) 
    shapeHess11.Assemble()

    shapeHess12 = BilinearForm(trialspace = fes, testspace = VEC)
    shapeHess12 += (G_pde.Diff(F, grad(V)) + G_pde.Diff(X, V)).Diff(gfu,w1)    
    shapeHess12.Assemble()

    shapeHess13 = BilinearForm(trialspace = fes, testspace = VEC)
    shapeHess13 += (G_pde.Diff(F, grad(V)) + G_pde.Diff(X, V)).Diff(gfp,q1)  
    shapeHess13.Assemble()

    av = gfV.vec.CreateVector()
    av.data = shapeHess11.mat * gfV.vec

    adsu = gfV.vec.CreateVector()
    adsu.data = shapeHess12.mat * gfdsu.vec

    adsp = gfV.vec.CreateVector()
    adsp.data = shapeHess13.mat * gfdsp.vec

    d2J =  InnerProduct(gfW.vec, av) + InnerProduct(gfW.vec, adsu) + InnerProduct(gfW.vec, adsp)
    
    input("End of Section 4.3. Press enter to continue.")
    ############## Section 5 #################
    G_f_0 = f * dx
    dJOmega_f_0 = LinearForm(VEC)
    dJOmega_f_0 += G_f_0.DiffShape(V)
    
    dJOmega_f.Assemble()
    dJOmega_f_0.Assemble()
    differenceVec = dJOmega_f.vec.CreateVector()
    differenceVec.data = dJOmega_f.vec - dJOmega_f_0.vec
    print("|dJOmega_f - dJOmega_f_0| = ", Norm(differenceVec) )
    
    d2JOmega_f_0 = BilinearForm(VEC)
    d2JOmega_f_0+=G_f_0.DiffShape(V).DiffShape(W)
    
    def Cost_0(u):
        return (u-ud)**2 * dx

    def Equation_0(u,w):
        return (grad(u) * grad(w) - f1*w) *dx

    G_pde_0 = Cost_0(gfu) + Equation_0(gfu, gfp)

    dJOmega_pde_0 = LinearForm(VEC)
    dJOmega_pde_0 += G_pde_0.DiffShape(V)
    
    shapeHess11_0 = BilinearForm(VEC)
    shapeHess11_0 += G_pde_0.DiffShape(W).DiffShape(V)
    
    G_f_bnd_0 = f * ds
    dJOmega_f_bnd_0 = LinearForm(VEC)
    dJOmega_f_bnd_0 += G_f_bnd_0.DiffShape(V)
    
    input("End of Section 5. Press enter to continue.")
    ############## Section 6.2.2 #################
    
    gfset = GridFunction(VEC)
    Draw(gfset, mesh, "gfset")
    SetVisualization (deformation=True)
    gfset.Set((X[0]*X[0]*X[1],X[1]*X[1]*X[0]))
    mesh.SetDeformation(gfset)
    Redraw()
    
    print(Integrate(1, mesh))
    mesh.UnsetDeformation()
    print(Integrate(1, mesh))
    
    def solvePDE():
        a.Assemble()
        L.Assemble()
        dCostdu.Assemble()

        inva = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky")    
        gfu.vec.data = inva * L.vec
        gfp.vec.data = -inva.T * dCostdu.vec
        
    def eps(u):
        return 1/2 * (grad(u)+grad(u).trans)

    aX = BilinearForm(VEC)
    W, V = VEC.TnT()    # define trial function W and test function V

    aX += InnerProduct(eps(W), eps(V))*dx + InnerProduct(W, V)*dx
    aX += 10 * (grad(W)[1,1] - grad(W)[0,0])*(grad(V)[1,1] - grad(V)[0,0])*dx
    aX += 10 * (grad(W)[1,0] + grad(W)[0,1])*(grad(V)[1,0] + grad(V)[0,1])*dx

    
    alpha = 1 
    alpha_incr_factor = 1.2
    gamma = 1e-4
    Nmax = 100
    epsilon = 1e-7

    isConverged = False
    gfset.Set((0,0))
    gfX = GridFunction(VEC)
    gfsetTemp = GridFunction(VEC)

    solvePDE()
    Jnew = Integrate(Cost(gfu), mesh)
    Jold = Jnew

    for k in range(Nmax):
        mesh.SetDeformation(gfset)
        aX.Assemble()
        dJOmega_pde.Assemble()
        invaX = aX.mat.Inverse(VEC.FreeDofs(), inverse="sparsecholesky")
        gfX.vec.data = invaX * dJOmega_pde.vec    
        currentNormGFX = Norm(gfX.vec)
        
        while True:
            if currentNormGFX < epsilon:
                isConverged = True
                break
            
            gfsetTemp.vec.data = gfset.vec - alpha * gfX.vec
            mesh.SetDeformation(gfsetTemp)
            solvePDE()
            Jnew = Integrate (Cost(gfu), mesh)
            mesh.UnsetDeformation()
            if Jnew < Jold - gamma * alpha * currentNormGFX**2 :
                Jold = Jnew
                gfset.vec.data = gfsetTemp.vec
                alpha *= alpha_incr_factor
                break
            else:
                alpha = alpha / 2
        Redraw(blocking=True)   
        
    def moveNGmesh2D(displ, mesh):
        for p in mesh.ngmesh.Points():
            v = displ(mesh(p[0],p[1]) )
            p[0] += v[0]
            p[1] += v[1]
        mesh.ngmesh.Update()    
        
        
    input("End of Section 6.2.2. Press enter to continue.")
    ############## Section 6.2.3 #################
        
    VEC2 = VectorH1(mesh, order=1, dirichlet = "circle") #auxiliary space for boundary conditions
    aX = BilinearForm(VEC)
    aX +=G_f_0.DiffShape(W).DiffShape(V)
    aX += 100 * InnerProduct(W, specialcf.tangential(2)) * InnerProduct(V, specialcf.tangential(2))*ds
    aX.Assemble()
    invAX = aX.mat.Inverse(~VEC2.FreeDofs(), inverse="umfpack")

    gfX_bnd = GridFunction(VEC)
    gfX_bnd.vec.data = invAX * dJOmega_f_0.vec
    
    def getExtension(gfX_bnd, freedofs, gfX_ext):
        u,v = VEC.TnT()
        aX_ext = BilinearForm(VEC)
        aX_ext += InnerProduct(grad(u)+grad(u).trans,grad(v))*dx+InnerProduct(u,v)*dx

        gfX_ext.Set(gfX_bnd)
        aX_ext.Assemble()

        r = gfX_bnd.vec.CreateVector()
        r.data = (-1) * aX_ext.mat * gfX_ext.vec

        gfX_ext.vec.data += aX_ext.mat.Inverse(freedofs=freedofs) * r
        
    getExtension(gfX_bnd, VEC2.FreeDofs(), gfX) 
    gfset.Set((0,0))
    gfset.vec.data = gfset.vec - 1 * gfX.vec
    
    input("End of Section 6.2.3. Press enter to continue.")
    ############## Section 6.2.4 #################
    
    X3 = FESpace([VEC, fes, fes])      
    PHI, u1, p1= X3.TrialFunction()
    PSI, uTest1, pTest1 = X3.TestFunction() 
    
    shapeHessLag3 = BilinearForm(X3)
    shapeHessLag3+= G_pde_0.DiffShape(PHI).DiffShape(PSI)        #block (1,1)
    shapeHessLag3 += G_pde_0.DiffShape(PSI).Diff(gfu,u1)         #block (1,2)
    shapeHessLag3 += G_pde_0.DiffShape(PSI).Diff(gfp,p1)         #block (1,3)
    shapeHessLag3 += G_pde_0.Diff(gfu, uTest1).DiffShape(PHI)    #block (2,1) 
    shapeHessLag3 += (G_pde_0.Diff(gfu, uTest1)).Diff(gfu, u1)   #block (2,2)
    shapeHessLag3 += (G_pde_0.Diff(gfu, uTest1)).Diff(gfp, p1)   #block (2,3)
    shapeHessLag3 += G_pde_0.Diff(gfp,pTest1).DiffShape(PHI)     #block (3,1)
    shapeHessLag3 += (G_pde_0.Diff(gfp,pTest1)).Diff(gfu, u1)    #block (3,2)
    
    shapeGradLag3 = LinearForm(X3)
    shapeGradLag3 += (-1) * G_pde_0.DiffShape(PSI)
    
    delta = 1
    shapeHessLag3 += delta * InnerProduct(PHI, specialcf.tangential(2)) * InnerProduct(PSI, specialcf.tangential(2))*ds
    
    # copy of VEC with Dirichlet boundary conditions on whole boundary:
    VEC2 = VectorH1(mesh, order = 1, dirichlet = ".*")    
    freeDofsCombined=BitArray(VEC2.ndof + 2*fes.ndof)
    for i in range(VEC2.ndof):
        freeDofsCombined[i] = not VEC2.FreeDofs()[i]
    for i in range(fes.ndof):
        freeDofsCombined[VEC2.ndof+i] = fes.FreeDofs()[i]
        freeDofsCombined[VEC2.ndof+fes.ndof+i] = fes.FreeDofs()[i]
    
    gfCombined3 = GridFunction(X3)
    shapeHessLag3.Assemble()
    shapeGradLag3.Assemble()
    gfCombined3.vec.data = shapeHessLag3.mat.Inverse(freedofs=freeDofsCombined, inverse="umfpack") * shapeGradLag3.vec
    
    Vtilde_bnd = GridFunction(VEC)
    Vtilde = GridFunction(VEC)
    Vtilde_bnd.vec.data = gfCombined3.components[0].vec
    getExtension(Vtilde_bnd, VEC2.FreeDofs(), Vtilde) 

    gfset.vec.data = gfset.vec + 1 * Vtilde.vec
    
    input("End of Section 6.2.4. Press enter to continue.")
    
    
    
    
    
    
    
    
    
else:       #if SURF == True
    ############## Section 4.4 #################
    from netgen.csg import *
    from netgen.meshing import *
    from ngsolve.internal import visoptions
    from ngsolve import *

    geo_surf = CSGeometry()
    sphere = Sphere(Pnt(0,0,0),1).bc("outer")
    geo_surf.Add(sphere)
    mesh_surf = Mesh(geo_surf.GenerateMesh(perfstepsend=MeshingStep.MESHSURFACE,optsteps2d=3,maxh=0.2))
    mesh_surf.Curve(3)
    fes_surf = H1(mesh_surf,order = 3)
    
    X = CoefficientFunction((x,y,z))
    func = CoefficientFunction(X[0]*X[1]*X[2])
    F = Id(3)
    tangDet = Det(F) * Norm( Inv(F).trans*specialcf.normal(3) )
    Bmat = (Id(3) - 1/Norm(Inv(F).trans*specialcf.normal(3))**2 * OuterProduct(Inv(F).trans*specialcf.normal(3), Inv(F).trans*specialcf.normal(3)) ) * Inv(F).trans

    def Equation_surf(u,w):
        return ( (Bmat*grad(u).Trace()) * (Bmat*grad(w).Trace()) + u*w - func * w) * tangDet * ds

    def Cost_surf(u):
        return u**2 * tangDet * ds
    
    #set up and solve state equation
    u_surf, w_surf = fes_surf.TnT()
    a = BilinearForm(fes_surf)
    a += Equation_surf(u_surf, w_surf)
    gfu_surf = GridFunction(fes_surf)
    solvers.Newton(a, gfu_surf, printing = False)
    Draw(gfu_surf, mesh_surf, "gfu_surf")
    
    #solve adjoint equation
    lfcost_surf = LinearForm(fes_surf)
    lfcost_surf += Cost_surf(gfu_surf).Diff(gfu_surf, w_surf)
    lfcost_surf.Assemble()
    inva = a.mat.Inverse(fes_surf.FreeDofs(), inverse="sparsecholesky")
    gfp_surf = GridFunction(fes_surf)
    gfp_surf.vec.data = -inva.T * lfcost_surf.vec
    Draw(gfp_surf, mesh_surf, "gfp_surf")
    
    G_surf = Cost_surf(gfu_surf) + Equation_surf(gfu_surf, gfp_surf)

    VEC3d = VectorH1(mesh_surf, order=1)
    V3d = VEC3d.TestFunction()
    dJOmega_surf = LinearForm(VEC3d)
    dJOmega_surf += G_surf.Diff(X, V3d) + G_surf.Diff(F, Grad(V3d).Trace())
    
    input("End of Section 4.4. Press enter to continue.")
    ############## Section 5 #################
    
    def Cost_surf_0(u):
        return u**2 * ds
    def Equation_surf_0(u,w):
        return  (grad(u).Trace()*grad(w).Trace() + u*w - func * w) * ds
    G_surf_0 = Cost_surf_0(gfu_surf) + Equation_surf_0(gfu_surf, gfp_surf)
    dJOmega_surf_0 = LinearForm(VEC3d)
    dJOmega_surf_0 += G_surf_0.DiffShape(V3d) 
    
    input("End of Section 5. Press enter to continue.")
