# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

import sys

import netgen.gui
import time

from ngsolve import *
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry

from ngsolve.internal import visoptions
from ngsolve.solvers import *
from math import pi, cos, sin

ngsglobals.msg_level = 1
visoptions.vecfunction='gfset'
SetVisualization (deformation=True)
X = CoefficientFunction((x,y,z))

exampleNumber = 7
exec(open("7a_geo.py").read())
exec(open("7b_pde.py").read())
exec(open("7c_cost.py").read())
    
def calcDescentDirectionAverage(gfX):
    mesh.UnsetDeformation()
    eta = [0.2]
    value = []
    x_coord = 0;
    N = 200
    for i in range(N+1):
        mySum = 0
        x_coord = 0.2+0.6*i/N
        eta.append(x_coord)
        for j in range(N):
            mySum = mySum + gfX(mesh(x_coord,j/N))*((j+1)/N-j/N)
        value.append(mySum)

    eta.append(0.8)
    spl = BSpline(2, eta, value)
    gfX_avg.Set( spl(x) )    

    mesh.SetDeformation(gfset);
    return gfX_avg;
    
VEC = VectorH1(mesh, order=order, dirichlet = "")
gfset = GridFunction(VEC)
gfset.Set((0,0))

# for averaged descent vector field
VEC1 = H1(mesh, order=4)    
PHIVEC, PSIVEC = VEC1.TnT()
gfX = GridFunction(VEC1);
gfX_avg = GridFunction(VEC1)

# define trial and test functions for state and adjoint problems
u, v = fes.TnT()
pTest1 = fes.TestFunction()

gfu = GridFunction(fes)
gfp = GridFunction(fes)

Draw (gfu, mesh, "state")
Draw (gfp, mesh, "adjoint")
Draw(gfX, mesh, "Vh(x,t)")
Draw(gfX_avg, mesh, "Vh_avg")
Draw(gfset,mesh,"gfset")

a = BilinearForm(fes)
a += Equation_vol(u, v)
G_vol = Cost_vol(gfu) + Equation_vol(gfu,gfp)

aV = BilinearForm(VEC1, symmetric=True)
aV += SymbolicBFI(InnerProduct(grad(PHIVEC),grad(PSIVEC)) + PHIVEC*PSIVEC);

fV = LinearForm(VEC1);
#fV += SymbolicLFI(grad(W)[0]*(gfu-ud)*(gfu-ud)-2*(gfu-ud)*ud.Diff(x)*W);
#fV += SymbolicLFI(grad(W)[0]*grad(gfu)[1]*gfp -grad(W)[0]*grad(gfu)[0]*grad(gfp)[0] );
#fV += SymbolicLFI(-grad(W)[0]*f*gfp - f.Diff(x)*W*gfp );

fV += G_vol.Diff(x, PSIVEC)
fV += G_vol.Diff(F11, grad(PSIVEC)[0])
fV += G_vol.Diff(F12, grad(PSIVEC)[1])

#define right hand side for adjoint equation
dCostdu = LinearForm(fes)
dCostdu += Cost_vol(gfu).Diff(gfu,pTest1)
##################### prepare output ###############
outputFolder = "output"
fileName = outputFolder + "/Jvals.m"
myFile = open(fileName,"w+")
myFile.write("Jvals_" + str(exampleNumber) + " = [")
myFile.close()
##################### prepare output ###############
    
input("press enter to start optimization")

t_optimization = time.time()

scale = 0.3
with TaskManager():
    for k in range(1000):
        mesh.SetDeformation(gfset)
        solvePDE()
        currentJ = Integrate (Cost_vol(gfu), mesh)
        print("it", k, 'cost', currentJ)

        aV.Assemble();
        fV.Assemble();
        gfX.vec.data = aV.mat.Inverse(VEC1.FreeDofs(),inverse="sparsecholesky") * fV.vec;
        gfX_avg = calcDescentDirectionAverage(gfX)
        ndof = len(gfX_avg.vec)
            
        gfset.vec.FV()[0:ndof] -= scale * gfX_avg.vec.FV()[0:ndof]
        
        currentNormGFX = Norm(gfX_avg.vec)
        print("------------------------------------------||gfX||", currentNormGFX )
        
        myFile = open(fileName,"a+")
        myFile.write(str(k) + " " + str(currentJ) + " " + str(currentNormGFX) + ";" )  
        myFile.close()
        
        Redraw(blocking=True)
        
        if currentNormGFX < 1e-7 or currentJ.real < 1e-8:
            break
    
    myFile = open(fileName,"a+")
    myFile.write("]; \n")
    myFile.close()

t_optimization = time.time() - t_optimization

myFile = open(fileName,"a+")
myFile.write("yyaxis left; plot(Jvals_" + str(exampleNumber) + "(:,1), Jvals_" + str(exampleNumber) +"(:,2)); ylabel('J') \n")
myFile.write("yyaxis right; plot( Jvals_" + str(exampleNumber) +"(:,1),  Jvals_" + str(exampleNumber) +"(:,3),'--'); ylabel('$\|\ nabla J\|$', 'Interpreter', 'latex') \n")
myFile.write("title('History of ...');  \n xlabel('Iterations') \n")
myFile.write("%  optimization time: " + str(t_optimization) + " seconds")
myFile.close()
breakpoint = input("Press any key to quit")
