# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

import sys

import netgen.gui
import time

from ngsolve import *
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry

from ngsolve.internal import visoptions
from ngsolve.solvers import *
from math import pi, cos, sin

ngsglobals.msg_level = 1
visoptions.vecfunction='gfset'
SetVisualization (deformation=True)
X = CoefficientFunction((x,y,z))

exampleNumber = 1
exec(open("1a_geo.py").read())
exec(open("1b_pde.py").read())
exec(open("1c_cost.py").read())
exec(open("1d_shapeopt.py").read())

############### Checking input... ############
def checkInput(stream=None):
    if stream is None:
        stream = sys.stdout
        
    stream.write("isComplex == " + str(isComplex) + "\n" )
    stream.write("isSurface == " + str(isSurface) + "\n" )
    #optimization parameters
    stream.write("newton == " + str(newton) +"\n")
    if newton == False:
        stream.write("lineSearch == " + str( lineSearch)  +"\n")
        if lineSearch == True:
            stream.write("armijo_param == " + str( armijo_param)  +"\n")
            stream.write("scaleIncreaseFactor == " + str( scaleIncreaseFactor)  +"\n")
        stream.write("scale_init == " + str( scale_init)  +"\n")
        stream.write("myAlpha == " + str(myAlpha)  +"\n")
    else:   #newton == True
        stream.write("regExt == " + str( regExt)  +"\n")
        stream.write("delta_regHess == " + str( delta_regHess)  +"\n")
        stream.write("nDampedSteps == " + str( nDampedSteps)  +"\n")
        
    stream.write("CR == " + str( CR)  +"\n")
    if CR == True:
        stream.write("gamma_CR == " + str( gamma_CR)  +"\n")
    
    stream.write("moveMesh == " + str(moveMesh)  +"\n")
    if moveMesh == True:
        stream.write("nMeshOpti == " + str(nMeshOpti)  +"\n")
        

    stream.write("iter_max == " + str( iter_max)  +"\n")
    stream.write("tol_scale == " + str( tol_scale)  +"\n")
    stream.write("tol_gfX == " + str( tol_gfX)  +"\n")
    stream.write("tol_J == " + str( tol_J)  +"\n")
    
    stream.write("\n ndofs == " + str(fes.ndof)  +"\n")
    stream.write("nVerts == " + str(mesh.nv)  +"\n")
    stream.write("nElems == " + str(mesh.ne)  +"\n")
    stream.write("nFaces == " + str(mesh.nface)  +"\n")
    stream.write("polyDegree == " + str(order)  +"\n")
    

#write parameters to file (create folder named "output" manually first)
outputFolder = "output"
fileName1 = outputFolder + "/params.txt"
myFile1 = open(fileName1,"w+")
checkInput(myFile1)
myFile1.close()

checkInput()        #writes parameters to console
###########################
def getExtension(gfX_bnd, freedofs, gfX_ext):
    u,v = VEC.TnT()
    aX_ext = BilinearForm(VEC)
    aX_ext += InnerProduct(grad(u)+grad(u).trans, grad(v))*dx# + InnerProduct(u,v)*dx

    gfX_ext.Set(gfX_bnd, BND)

    aX_ext.Assemble()

    r = gfX_bnd.vec.CreateVector()
    r.data = (-1) * aX_ext.mat * gfX_bnd.vec

    gfX_ext.vec.data += aX_ext.mat.Inverse(freedofs=freedofs) * r
###########################################
def moveNGmesh(displ, mesh):
    for p in mesh.ngmesh.Points():
        v = displ(mesh(p[0],p[1]))
        p[0] += v[0]
        p[1] += v[1]
    mesh.ngmesh.Update()    
###########################################
VEC = VectorH1(mesh, order=order, dirichlet = dirichletOpti)

freeDofsCombined=BitArray(VEC.ndof)
for i in range(VEC.ndof):
    freeDofsCombined[i] = not VEC.FreeDofs()[i]

# for first order method
PHIVEC = VEC.TrialFunction()
PSIVEC = VEC.TestFunction()

# define trial and test functions for state and adjoint problems
u, v = fes.TnT()
pTest1 = fes.TestFunction()

gfset = GridFunction(VEC)
gfsetTemp = GridFunction(VEC)
if mesh.dim==2:
    gfset.Set((0,0))
elif mesh.dim == 3:
    gfset.Set((0,0,0))

gfX = GridFunction(VEC)
gfX_bnd = GridFunction(VEC)
gfu = GridFunction(fes)
gfp = GridFunction(fes)

G_vol = Cost_vol(gfu)
G_bnd = Cost_bnd(gfu)

Draw(gfX,mesh,"gfX")
Draw(gfX_bnd,mesh,"gfX_bnd")
Draw(gfset,mesh,"gfset")

if newton == True:
    shapeHessLag = BilinearForm(VEC)
    
    shapeHessLag+= ( G_vol.Diff(F,grad(PHIVEC)) +  G_vol.Diff(X, PHIVEC) ).Diff(F,grad(PSIVEC))
    shapeHessLag+= ( G_vol.Diff(F,grad(PHIVEC)) +  G_vol.Diff(X, PHIVEC) ).Diff(X, PSIVEC)
    
    shapeHessLag+= ( G_bnd.Diff(F,grad(PHIVEC).Trace()) +  G_bnd.Diff(X, PHIVEC) ).Diff(F,grad(PSIVEC).Trace())
    shapeHessLag+= ( G_bnd.Diff(F,grad(PHIVEC).Trace()) +  G_bnd.Diff(X, PHIVEC) ).Diff(X, PSIVEC)
    
    if regExt == True:
        shapeHessLag += delta_regHess * InnerProduct(PHIVEC, specialcf.tangential(mesh.dim)) * InnerProduct(PSIVEC, specialcf.tangential(mesh.dim))*ds
        print("delta_regHess", delta_regHess)
    else:
        shapeHessLag += delta_regHess * InnerProduct(grad(PHIVEC), grad(PSIVEC)) * dx
        shapeHessLag += delta_regHess * InnerProduct(PHIVEC, PSIVEC) * dx
        
    if CR == True:
        shapeHessLag += gamma_CR * (grad(PHIVEC)[0,0] - grad(PHIVEC)[1,1])*(grad(PSIVEC)[0,0] - grad(PSIVEC)[1,1]) *dx
        shapeHessLag += gamma_CR * (grad(PHIVEC)[1,0] + grad(PHIVEC)[0,1])*(grad(PSIVEC)[1,0] + grad(PSIVEC)[0,1]) *dx
    
    cHess = Preconditioner(shapeHessLag, "direct")
    #####
    shapeGradLag = LinearForm(VEC)
    shapeGradLag += G_vol.Diff(F, grad(PSIVEC)) + G_vol.Diff(X, PSIVEC)
    shapeGradLag += G_bnd.Diff(F, grad(PSIVEC).Trace()) + G_bnd.Diff(X, PSIVEC)
        
else:   #i.e. newton == False
    aX = BilinearForm(VEC)
    if ela:
        aX += myAlpha * InnerProduct(0.5* (grad(PHIVEC).trans + grad(PHIVEC)), grad(PSIVEC))*dx + InnerProduct(PHIVEC,PSIVEC)*dx
    else:
        aX += myAlpha * InnerProduct(grad(PHIVEC), grad(PSIVEC))*dx + InnerProduct(PHIVEC,PSIVEC)*dx
    if CR == True:
        aX += gamma_CR * (grad(PHIVEC)[0,0] - grad(PHIVEC)[1,1])*(grad(PSIVEC)[0,0] - grad(PSIVEC)[1,1]) *dx
        aX += gamma_CR * (grad(PHIVEC)[1,0] + grad(PHIVEC)[0,1])*(grad(PSIVEC)[1,0] + grad(PSIVEC)[0,1]) *dx
    
    
    fX = LinearForm(VEC)    
    #fX += G_vol.DiffShape(PSIVEC)
    #fX += G_bnd.DiffShape(PSIVEC)
    
    fX += G_vol.Diff(F, grad(PSIVEC))
    fX += G_vol.Diff(X, PSIVEC)
    
    fX += G_bnd.Diff(F, grad(PSIVEC).Trace())
    fX += G_bnd.Diff(X, PSIVEC)

    aX.Assemble()
    invaX = aX.mat.Inverse(VEC.FreeDofs(), inverse="sparsecholesky")
    
##################### prepare output ###############
fileName = outputFolder + "/Jvals.m"
myFile = open(fileName,"w+")
myFile.write("Jvals_" + str(exampleNumber) + " = [")
myFile.close()
##################### prepare output ###############
    
input("press enter to start optimization")

t_optimization = time.time()
if newton == True or lineSearch == False:
    if newton == True:
        scale = 1
    else:
        scale = scale_init
    with TaskManager():
        for k in range(iter_max):
            if moveMesh == False:
                mesh.SetDeformation(gfset)
            currentJ = Integrate (Cost_vol(gfu)+Cost_bnd(gfu), mesh)
            print("it", k, 'cost', currentJ)

            if newton == True:
                shapeGradLag.Assemble()
                shapeHessLag.Assemble()
                if regExt:
                    gfX_bnd.vec.data = shapeHessLag.mat.Inverse(freedofs=freeDofsCombined, inverse="umfpack") * shapeGradLag.vec
                    getExtension(gfX_bnd, VEC.FreeDofs(), gfX)
                else:
                    gfX.vec.data = shapeHessLag.mat.Inverse(VEC.FreeDofs(), inverse="umfpack") * shapeGradLag.vec
            else:
                aX.Assemble()
                fX.Assemble()
                invaX.Update()
                gfX.vec.data = invaX * fX.vec
                
            #damped newton for regExt            
            if newton == True and nDampedSteps > 0 and k < nDampedSteps:
                scale = (k+1) / nDampedSteps
            
            if moveMesh == True:
                moveNGmesh(-scale * gfX, mesh)
                VEC.Update()
            else:
                gfset.vec.data -= scale * gfX.vec
            
            currentNormGFX = Norm(gfX.vec)
            print("------------------------------------------||gfX||", currentNormGFX )
            
            myFile = open(fileName,"a+")
            myFile.write(str(k) + " " + str(currentJ) + " " + str(currentNormGFX) + ";" )  
            myFile.close()
            
            Redraw(blocking=True)
            
            if scale < tol_scale or currentNormGFX < tol_gfX or currentJ.real < tol_J:
                break
        
        myFile = open(fileName,"a+")
        myFile.write("]; \n")
        myFile.close()

else:   #newton == False and lineSeach == True
    with TaskManager():        
        isConverged = False
        if moveMesh == False:
            mesh.SetDeformation(gfset)
        Jnew = Integrate (Cost_vol(gfu)+Cost_bnd(gfu), mesh)
        if moveMesh == False:
            mesh.UnsetDeformation()
        Jold = Jnew
        print("it init", 'cost', Jnew )
        
        scale = scale_init 
        for k in range(iter_max):
            if moveMesh == False:
                mesh.SetDeformation(gfset)
            aX.Assemble()
            fX.Assemble()
            invaX = aX.mat.Inverse(VEC.FreeDofs(), inverse="sparsecholesky")
            gfX.vec.data = invaX * fX.vec
            
            currentNormGFX = Norm(gfX.vec)
            myFile = open(fileName,"a+")
            myFile.write(str(k) + " " + str(Jnew) + " " + str(currentNormGFX) + ";" )  
            myFile.close()
            
            while True:
                if scale < tol_scale or currentNormGFX < tol_gfX or Jnew.real < tol_J:
                    isConverged = True
                    break
            
                if moveMesh == False:
                    gfsetTemp.vec.data = gfset.vec - scale * gfX.vec
                    mesh.SetDeformation(gfsetTemp)
                else:
                    moveNGmesh(-scale * gfX, mesh)
                    VEC.Update()

                Jnew = Integrate (Cost_vol(gfu)+Cost_bnd(gfu), mesh)
                if moveMesh == False:
                    mesh.UnsetDeformation()
                    
                print("Jnew = ", Jnew.real, "condition = ", Jold.real - armijo_param*scale * currentNormGFX**2)
                if Jnew.real < Jold.real - armijo_param*scale * currentNormGFX**2 :
                    Jold = Jnew
                    print("----------it", k, 'scale', scale, 'cost', Jnew )
                    if moveMesh == False:
                        gfset.vec.data = gfsetTemp.vec
                    scale *= scaleIncreaseFactor
                    break
                else:
                    if moveMesh == True:
                        moveNGmesh(scale * gfX, mesh)
                        VEC.Update()
                        fes.Update()
                    scale = scale / 2
                
            print("------------------------------------------||gfX||", currentNormGFX )
            print("isConverged", isConverged)
            Redraw(blocking=True)
            if isConverged== True:
                print("converged with J = ", Jnew.real, ", ||gfX||=", currentNormGFX, ", scale = ", scale)
                print("Thank you for using NGShapeOpt")
                break
        myFile = open(fileName,"a+")
        myFile.write("]; \n")
        myFile.close()

t_optimization = time.time() - t_optimization

myFile = open(fileName,"a+")
myFile.write("yyaxis left; plot(Jvals_" + str(exampleNumber) + "(:,1), Jvals_" + str(exampleNumber) +"(:,2)); ylabel('J') \n")
myFile.write("yyaxis right; plot( Jvals_" + str(exampleNumber) +"(:,1),  Jvals_" + str(exampleNumber) +"(:,3),'--'); ylabel('$\|\ nabla J\|$', 'Interpreter', 'latex') \n")
myFile.write("title('History of ...');  \n xlabel('Iterations') \n")
myFile.write("%  optimization time: " + str(t_optimization) + " seconds")
myFile.close()
breakpoint = input("Press any key to quit")
