# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

import netgen.gui
from ngsolve import *
from netgen.geom2d import SplineGeometry

testall = True
stop    = False

geo = SplineGeometry()
geo.AddCircle((0,0), r = 0.3)
ngmesh = geo.GenerateMesh(maxh = 0.01) 
mesh = Mesh(ngmesh)
mesh.Curve(2)

n = specialcf.normal(2)
tang = specialcf.tangential(2)

VEC = VectorH1(mesh, order=2)
# vector space for shape gradient
VEC = VectorH1(mesh, order=2)
# grid function for deformation field
gfX = GridFunction(VEC)
gfX_t = GridFunction(VEC)
# Test and trial functions
PHI, PSI = VEC.TnT()

#gfX.Set((x,y))
gfX.Set((x**2*y*exp(y),y**2*x*exp(x)))

X = CoefficientFunction( (x,y) )
F = Id(2)

uD = CoefficientFunction( 2*X[1]*(1-X[1])+2*X[0]*(1-X[0]) )
grad_uD = CoefficientFunction( (uD.Diff(x), uD.Diff(y)) )

def Cost(u):
    return (u-uD)**2

vuD = CoefficientFunction( (uD,uD) )
vgrad_uD = CoefficientFunction( (grad_uD,grad_uD), dims=(2,2) )
def CostV(u):
    return (u-vuD)**2

def Test(G0, G, dJOmegaTest, G0bnd=None, Gbnd=None):
    #semi automatic
    dJOmegaTestSA = LinearForm(VEC)
    dJOmegaTestSA += G.Diff(F, Grad(PSI))
    dJOmegaTestSA += G.Diff(X, PSI)
    if Gbnd:
        dJOmegaTestSA += Gbnd.Diff(F, Grad(PSI).Trace())
        dJOmegaTestSA += Gbnd.Diff(X, PSI)
    ddJOmegaTestSA = BilinearForm(VEC)
    ddJOmegaTestSA += (G.Diff(F, Grad(PSI)) + G.Diff(X, PSI)).Diff(F, Grad(PHI)).Compile()
    ddJOmegaTestSA += (G.Diff(F, Grad(PSI)) + G.Diff(X, PSI)).Diff(X, PHI).Compile()
    if Gbnd:
        ddJOmegaTestSA += (Gbnd.Diff(F, Grad(PSI).Trace()) + Gbnd.Diff(X, PSI)).Diff(F, Grad(PHI).Trace()).Compile()
        ddJOmegaTestSA += (Gbnd.Diff(F, Grad(PSI).Trace()) + Gbnd.Diff(X, PSI)).Diff(X, PHI).Compile()
        
    #full automatic
    dJOmegaTestFA = LinearForm(VEC)
    dJOmegaTestFA += G0.DiffShape(PSI)
    if G0bnd:
        dJOmegaTestFA += G0bnd.DiffShape(PSI)
    ddJOmegaTestFA = BilinearForm(VEC)
    ddJOmegaTestFA += G0.DiffShape(PSI).DiffShape(PHI).Compile()
    if G0bnd:
        ddJOmegaTestFA += G0bnd.DiffShape(PSI).DiffShape(PHI).Compile()

    t = 1
    J0 = Integrate(G0, mesh)
    if G0bnd:
        J0 += Integrate(G0bnd, mesh)
    print("J0", J0 )

    delta1 = [0,0,0]
    delta2 = [0,0]
    delta1Prev = [0,0,0]
    delta2Prev = [0,0]
    dJOmegaTest.Assemble()
    dJOmegaTestSA.Assemble()
    dJOmegaTestFA.Assemble()
    ddJOmegaTestSA.Assemble()
    ddJOmegaTestFA.Assemble()

    tmp1 = dJOmegaTest.vec.CreateVector()
    tmp2 = dJOmegaTest.vec.CreateVector()
    tmp1.data = ddJOmegaTestSA.mat*gfX.vec
    tmp2.data = ddJOmegaTestFA.mat*gfX.vec
    
    for i in range(10):
        for j in range(3):
            delta1Prev[j] = delta1[j]
        for j in range(2):
            delta2Prev[j] = delta2[j]
        t = t/2
        gfX_t.vec.data = t*gfX.vec
        print("t = ", t)
        # deform current domain with gfX_t
        mesh.SetDeformation(gfX_t)
        Jt = Integrate(G0, mesh)
        if G0bnd:
            Jt += Integrate(G0bnd, mesh)
        print("Jt =  ",  Jt)
        mesh.UnsetDeformation()
        delta1[0] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTest.vec,gfX.vec))
        delta1[1] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestSA.vec,gfX.vec))
        delta1[2] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestFA.vec,gfX.vec))
        delta2[0] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestSA.vec,gfX.vec) - t**2/2*InnerProduct(tmp1,gfX.vec))
        delta2[1] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestFA.vec,gfX.vec) - t**2/2*InnerProduct(tmp2,gfX.vec))
        print( " delta1 = ", delta1)
        print( " delta2 = ", delta2)
        print( "--------------------------delta1Prev / delta1 = ",round(delta1Prev[0] / delta1[0],3),round(delta1Prev[1] / delta1[1],3),round(delta1Prev[2] / delta1[2],3))
        print( "--------------------------delta2Prev / delta2 = ",round(delta2Prev[0] / delta2[0],3),round(delta2Prev[1] / delta2[1],3))
    return



######################## Basic ######################
G0 = uD * dx
G  = uD * Det(F) * dx

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (div(PSI) * uD + grad_uD * PSI) * dx
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("1. press enter to continue")


######################## H1 ######################
fes = H1(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( x*y )
gfp.Set( 4*x - 2*y**2 )

f = grad(gfu)*grad(gfp)
G0 = (Cost(gfu) + f)* dx
G  = ((Inv(F).trans*grad(gfu))*(Inv(F).trans*grad(gfp)) + Cost(gfu)) * Det(F) * dx

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (div(PSI) * (f + Cost(gfu))  - (Grad(PSI)+Grad(PSI).trans)*grad(gfu)*grad(gfp) - 2*(gfu-uD)*grad_uD*PSI) * dx
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("2. press enter to continue")

######################## VectorH1 ######################
fes = VectorH1(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( CoefficientFunction( (x*y, -x+y) ) )
gfp.Set( CoefficientFunction( (x - y**2, -x*2*y) ) )

f = InnerProduct(grad(gfu),grad(gfp)) + gfu*gfp + Trace(grad(gfu))
G0 = (CostV(gfu) + f)* dx
G  = (InnerProduct(grad(gfu)*Inv(F),grad(gfp)*Inv(F)) + gfu*gfp + Trace(grad(gfu)*Inv(F))  + CostV(gfu)) * Det(F) * dx

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (div(PSI) * (f + CostV(gfu))  - InnerProduct(grad(gfu)*Grad(PSI),grad(gfp)) - InnerProduct(grad(gfu),grad(gfp)*Grad(PSI)) -Trace(grad(gfu)*Grad(PSI)) - 2*(gfu-vuD)*(vgrad_uD*PSI)) * dx
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("3. press enter to continue")


######################## HCurl ######################
fes = HCurl(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( CoefficientFunction( (x*y, -x+y) ) )
gfp.Set( CoefficientFunction( (x - y**2, -x*2*y) ) )

f = curl(gfu)*curl(gfp) + gfu*gfp 
G0 = (CostV(gfu) + f)* dx
G  = (InnerProduct(Inv(F).trans*gfu,Inv(F).trans*gfp) + 1/Det(F)**2*(curl(gfu))*(curl(gfp)) + CostV(Inv(F).trans*gfu)) * Det(F) * dx

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (div(PSI) * (f + CostV(gfu)) -((Grad(PSI)+Grad(PSI).trans)*gfu)*gfp - 2*div(PSI)*curl(gfu)*curl(gfp) - 2*(gfu-vuD)*(vgrad_uD*PSI) - 2*(gfu-vuD)*(Grad(PSI).trans*gfu)) * dx
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("4. press enter to continue")


######################## HDiv ######################
fes = HDiv(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( CoefficientFunction( (x*y, -x+y) ) )
gfp.Set( CoefficientFunction( (x - y**2, -x*2*y) ) )

f = gfu*gfp + div(gfu)*div(gfp)
G0 = (CostV(gfu) + f)* dx
G  = ( 1/Det(F)**2*(F*gfu)*(F*gfp) + 1/Det(F)**2*div(gfu)*div(gfp) + CostV(1/Det(F)*F*gfu)) * Det(F) * dx

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (div(PSI) * (f + CostV(gfu)) -2*div(PSI)*div(gfu)*div(gfp) -2*div(PSI)*gfu*gfp +((Grad(PSI)+Grad(PSI).trans)*gfu)*gfp - 2*(gfu-vuD)*(vgrad_uD*PSI) + 2*(gfu-vuD)*(-div(PSI)*gfu + Grad(PSI)*gfu)) * dx
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("5. press enter to continue")


######################## specialcf/BND ######################
fes = VectorH1(mesh, order=2)
gfu = GridFunction(fes)
gfu.Set( CoefficientFunction( (x*3*y, -x*(x+y)) ) )

f = 1000*(gfu.Trace()*n + gfu.Trace()*tang)
G0 = CostV(gfu) * dx
G0bnd = f*ds
G  = CostV(gfu) * Det(F) * dx
Gbnd = 1000*(1/Norm(Inv(F).trans*n)*gfu.Trace()*(Inv(F).trans*n) + 1/Norm(F*tang)*gfu.Trace()*(F*tang))*Det(F)*Norm(Inv(F).trans*n)*ds

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (div(PSI)*CostV(gfu)- 2*(gfu-vuD)*(vgrad_uD*PSI))*dx
dJOmegaTest += (Trace(Grad(PSI).Trace())*f + 1000*(gfu.Trace()*(-Grad(PSI).Trace().trans*n + ((Grad(PSI).Trace()*n)*n)*n) + gfu.Trace()*(Grad(PSI).Trace()*tang-((Grad(PSI).Trace()*tang)*tang)*tang)))*ds
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest,G0bnd,Gbnd)
    if stop:
        input("6. press enter to continue")

