# Supplementary material to the manuscript entitled
#       "Fully and Semi-Automated Shape Differentiation in NGSolve"
# by the authors
#       P. Gangl, K. Sturm, M. Neunteufel, J. Schoeberl
# submitted to the journal
#        Structural and Multidisciplinary Optimization
# Corresponding author: 
#       Peter Gangl, Institute of Applied Mathematics, 
#       Graz University of Technology, Graz, Austria
#       gangl@math.tugraz.at

from netgen.csg import *
from netgen.meshing import *
from ngsolve.internal import visoptions
from ngsolve import *

testall = True
stop    = False

geo = CSGeometry()
sphere = Sphere(Pnt(0,0,0),1).bc("outer")
geo.Add(sphere)
mesh = Mesh(geo.GenerateMesh(perfstepsend=MeshingStep.MESHSURFACE,optsteps2d=3,maxh=0.2))
mesh.Curve(3)

n = specialcf.normal(3)
n.dims =(3,1)
tang = specialcf.tangential(3)

VEC = VectorH1(mesh, order=2)
# vector space for shape gradient
VEC = VectorH1(mesh, order=2)
# grid function for deformation field
gfX = GridFunction(VEC)
gfX_t = GridFunction(VEC)
# Test and trial functions
PHI, PSI = VEC.TnT()

#gfX.Set((x,y,z), definedon=mesh.Boundaries(".*"))
gfX.Set((x**2*z*exp(y),y**2*x*exp(z),z**2*y*exp(x)), definedon=mesh.Boundaries(".*"))

X = CoefficientFunction( (x,y,z) )
F = Id(3)

tangDet = Det(F)*Norm(Inv(F).trans*n)
Bmat = (Id(3)-1/Norm(Inv(F).trans*n)**2*OuterProduct(Inv(F).trans*n, Inv(F).trans*n))*Inv(F).trans
Amat = Inv(F)*(Id(3)-1/Norm(Inv(F).trans*n)**2*OuterProduct(Inv(F).trans*n, Inv(F).trans*n))

uD = CoefficientFunction( 2*X[1]*(1-X[1])+2*X[0]*(1-X[0]) + 2*X[2]*(1-X[2]) )
grad_uD = CoefficientFunction( (uD.Diff(x), uD.Diff(y), uD.Diff(z)) )
hesse_uD = CoefficientFunction( (grad_uD[0].Diff(x),grad_uD[0].Diff(y),grad_uD[0].Diff(z),
                                 grad_uD[1].Diff(x),grad_uD[1].Diff(y),grad_uD[1].Diff(z),
                                 grad_uD[2].Diff(x),grad_uD[2].Diff(y),grad_uD[2].Diff(z)), dims=(3,3) )

def Cost(u):
    return (u-uD)**2

vuD = CoefficientFunction( (uD,uD, uD) )
vgrad_uD = CoefficientFunction( (grad_uD,grad_uD,grad_uD), dims=(3,3) )
def CostV(u):
    return (u-vuD)**2

def Test(G0, G, dJOmegaTest, ddJOmegaTest=None, G0bbnd=None, Gbbnd=None):
    #semi automatic
    dJOmegaTestSA = LinearForm(VEC)
    dJOmegaTestSA += G.Diff(F, Grad(PSI).Trace())
    dJOmegaTestSA += G.Diff(X, PSI)
    if Gbbnd:
        dJOmegaTestSA += Gbbnd.Diff(F, Grad(PSI).Trace().Trace())
        dJOmegaTestSA += Gbbnd.Diff(X, PSI)
        
    ddJOmegaTestSA = BilinearForm(VEC)
    ddJOmegaTestSA += (G.Diff(F, Grad(PSI).Trace()) + G.Diff(X, PSI)).Diff(F, Grad(PHI).Trace())
    ddJOmegaTestSA += (G.Diff(F, Grad(PSI).Trace()) + G.Diff(X, PSI)).Diff(X, PHI)
    if Gbbnd:
        ddJOmegaTestSA += (Gbbnd.Diff(F, Grad(PSI).Trace().Trace()) + Gbnd.Diff(X, PSI)).Diff(F, Grad(PHI).Trace().Trace())
        ddJOmegaTestSA += (Gbbnd.Diff(F, Grad(PSI).Trace().Trace()) + Gbnd.Diff(X, PSI)).Diff(X, PHI)
    
    #full automatic
    dJOmegaTestFA = LinearForm(VEC)
    dJOmegaTestFA += G0.DiffShape(PSI)
    if G0bbnd:
        dJOmegaTestFA += G0bbnd.DiffShape(PSI)
        
    ddJOmegaTestFA = BilinearForm(VEC)
    ddJOmegaTestFA += G0.DiffShape(PSI).DiffShape(PHI)
    if G0bbnd:
        ddJOmegaTestFA += G0bbnd.DiffShape(PSI).DiffShape(PHI)
   
    t = 1
    J0 = Integrate(G0, mesh)
    if G0bbnd:
        J0 += Integrate(G0bnd, mesh)
    print("J0", J0 )

    delta1 = [0,0,0]
    delta2 = [0,0,0]
    delta1Prev = [0,0,0]
    delta2Prev = [0,0,0]
    dJOmegaTest.Assemble()
    if ddJOmegaTest:
        ddJOmegaTest.Assemble()
    dJOmegaTestSA.Assemble()
    dJOmegaTestFA.Assemble()
    ddJOmegaTestSA.Assemble()
    ddJOmegaTestFA.Assemble()
    
    tmp0 = dJOmegaTest.vec.CreateVector()
    tmp1 = dJOmegaTest.vec.CreateVector()
    tmp2 = dJOmegaTest.vec.CreateVector()
    if ddJOmegaTest:
        tmp0.data = ddJOmegaTest.mat*gfX.vec
    tmp1.data = ddJOmegaTestSA.mat*gfX.vec
    tmp2.data = ddJOmegaTestFA.mat*gfX.vec
    
    for i in range(10):
        for j in range(3):
            delta1Prev[j] = delta1[j]
        for j in range(3):
            delta2Prev[j] = delta2[j]
        t = t/2
        gfX_t.vec.data = t*gfX.vec
        print("t = ", t)
        # deform current domain with gfX_t
        mesh.SetDeformation(gfX_t)
        Jt = Integrate(G0, mesh)
        if G0bbnd:
            Jt += Integrate(G0bbnd, mesh)
        print("Jt =  ",  Jt)
        mesh.UnsetDeformation()
        delta1[0] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTest.vec,gfX.vec))
        delta1[1] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestSA.vec,gfX.vec))
        delta1[2] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestFA.vec,gfX.vec))
        if ddJOmegaTest:
            delta2[0] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTest.vec,gfX.vec) - t**2/2*InnerProduct(tmp0,gfX.vec))
        else:
            delta2[0] = 1
        delta2[1] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestSA.vec,gfX.vec) - t**2/2*InnerProduct(tmp1,gfX.vec))
        delta2[2] = abs(Jt - J0 -  t*InnerProduct(dJOmegaTestFA.vec,gfX.vec) - t**2/2*InnerProduct(tmp2,gfX.vec))
        print( " delta1 = ", delta1)
        print( " delta2 = ", delta2)
        print( "--------------------------delta1Prev / delta1 = ",round(delta1Prev[0] / delta1[0],3),round(delta1Prev[1] / delta1[1],3),round(delta1Prev[2] / delta1[2],3))
        print( "--------------------------delta2Prev / delta2 = ",round(delta2Prev[0] / delta2[0],3),round(delta2Prev[1] / delta2[1],3),round(delta2Prev[2] / delta2[2],3))
    return



######################## Basic ######################
G0 = uD * ds
G  = uD * tangDet * ds

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (Trace(Grad(PSI).Trace()) * uD + grad_uD * PSI) * ds
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("1. press enter to continue")


######################## H1 ######################
fes = H1(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( x*y, definedon=mesh.Boundaries(".*") )
gfp.Set( z - x**2+exp(y), definedon=mesh.Boundaries(".*") )

f = grad(gfu).Trace()*grad(gfp).Trace()
G0 = (Cost(gfu) + f)* ds
G  = ((Bmat*grad(gfu).Trace())*(Bmat*grad(gfp).Trace()) + Cost(gfu)) * tangDet * ds

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (Trace(Grad(PSI).Trace()) * (f + Cost(gfu))  - (Grad(PSI).Trace()+Grad(PSI).Trace().trans)*grad(gfu).Trace()*grad(gfp).Trace() - 2*(gfu-uD)*grad_uD*PSI) * ds
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("2. press enter to continue")

######################## VectorH1 ######################
fes = VectorH1(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( CoefficientFunction( (x*y, -x+y, z*x*y) ), definedon=mesh.Boundaries(".*") )
gfp.Set( CoefficientFunction( (z - x**2+exp(y), x*exp(y), exp(z)) ), definedon=mesh.Boundaries(".*") )

f = gfu.Trace()*gfp.Trace() + InnerProduct(grad(gfu).Trace(),grad(gfp).Trace())
G0 = (CostV(gfu.Trace()) + f)* ds
G  = ( gfu.Trace()*gfp.Trace() + InnerProduct(grad(gfu).Trace()*Amat,grad(gfp).Trace()*Amat)  + CostV(gfu.Trace())) * tangDet * ds

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (Trace(Grad(PSI).Trace()) * (f + CostV(gfu)) -InnerProduct(grad(gfu).Trace()*Grad(PSI).Trace(),Grad(gfp).Trace()) -InnerProduct(grad(gfu).Trace(),Grad(gfp).Trace()*Grad(PSI).Trace())  - 2*(gfu-vuD)*(vgrad_uD*PSI)) * ds
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("3. press enter to continue")


######################## HCurl ######################
fes = HCurl(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( CoefficientFunction( (x*y, -x+y, z*x*y) ), definedon=mesh.Boundaries(".*") )
gfp.Set( CoefficientFunction( (z - x**2+exp(y), x*exp(y), exp(z)) ), definedon=mesh.Boundaries(".*") )

f =  gfu.Trace()*gfp.Trace() + curl(gfu).Trace()*curl(gfp).Trace()
G0 = (CostV(gfu.Trace()) + f)* ds
G  = (InnerProduct(Bmat*gfu.Trace(),Bmat*gfp.Trace()) + 1/tangDet**2*(F*curl(gfu).Trace())*(F*curl(gfp).Trace()) + CostV(Bmat*gfu.Trace())) * tangDet * ds

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (Trace(Grad(PSI).Trace()) * (f + CostV(gfu.Trace())) - (Grad(PSI).Trace()+Grad(PSI).Trace().trans)*gfu.Trace()*gfp.Trace() -2*Trace(Grad(PSI).Trace())*curl(gfu).Trace()*curl(gfp).Trace() +((Grad(PSI).Trace()+Grad(PSI).Trace().trans)*curl(gfu).Trace())*curl(gfp).Trace() - 2*(gfu.Trace()-vuD)*(vgrad_uD*PSI) + 2*(gfu.Trace()-vuD)*((2*Sym((n*n.trans)*Grad(PSI).Trace())-Grad(PSI).Trace().trans)*gfu.Trace())) * ds
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("4. press enter to continue")


######################## HDiv ######################
fes = HDivSurface(mesh, order=1)
gfu = GridFunction(fes)
gfp = GridFunction(fes)
gfu.Set( CoefficientFunction( (x*y, -x+y, z*x*y) ), definedon=mesh.Boundaries(".*") )
gfp.Set( CoefficientFunction( (z - x**2+exp(y), x*exp(y), exp(z)) ), definedon=mesh.Boundaries(".*") )

f = gfu.Trace()*gfp.Trace() + div(gfu).Trace()
G0 = (CostV(gfu.Trace()) + f)* ds
G  = ( 1/tangDet**2*(F*gfu.Trace())*(F*gfp.Trace())  +1/tangDet*div(gfu).Trace()  + CostV(1/tangDet*F*gfu.Trace())) * tangDet * ds

dJOmegaTest = LinearForm(VEC)
dJOmegaTest += (Trace(Grad(PSI).Trace()) * (f + CostV(gfu.Trace())) -Trace(Grad(PSI).Trace())*div(gfu).Trace() -2*Trace(Grad(PSI).Trace())*gfu.Trace()*gfp.Trace() +((Grad(PSI).Trace()+Grad(PSI).Trace().trans)*gfu)*gfp - 2*(gfu.Trace()-vuD)*(vgrad_uD*PSI) + 2*(gfu-vuD)*(-Trace(Grad(PSI).Trace())*gfu.Trace() + Grad(PSI).Trace()*gfu)) * ds
if testall:
    with TaskManager():
        Test(G0, G, dJOmegaTest)
    if stop:
        input("5. press enter to continue")
